#ifndef __SYSTEM_H
#define __SYSTEM_H

# define u32int unsigned int
# define u16int unsigned short
# define u8int unsigned char

struct regs
{
    unsigned int gs, fs, es, ds;
    unsigned int edi, esi, ebp, esp, ebx, edx, ecx, eax;
    unsigned int int_no, err_code;
    unsigned int eip, cs, eflags, useresp, ss;    
};


/* To read the input string typed by the user */
char *keyboardBuffer;
int bufferFlag;

/* MAIN.C */
extern unsigned char *memcpy(void *dest, const void *src, int count);
extern unsigned char *memset(void *dest, char val, int count);
extern unsigned short* memsetw(unsigned short *dest, unsigned short val, int count);
extern int strlen(const char* str);
extern unsigned char inportb(unsigned short _port);
extern void outportb(unsigned short _port, unsigned char _data);

/* SCRN.C */
extern void init_video(void);
extern void settextcolor(unsigned char forecolor, unsigned char backcolor);
extern void puts(unsigned char* text);
extern void putch(unsigned char c);
extern void putNumber(int str);
extern void puthex(u32int n);
extern void cls();
extern void move_scr(void);
extern void scroll(void);

/* GDT.C */
extern void gdt_install();
extern void gdt_set_gate(int num, unsigned long base, unsigned long limit, unsigned char access, unsigned char gram);

/* IDT.C */
extern void idt_install();
extern void idt_set_gate(unsigned char num, unsigned long base, unsigned short sel, unsigned char flags);

/* ISRS.C */
extern void isrs_install();
extern void register_interrupt_handler(int n, void(*handler)(struct regs *r));
extern void fault_handler(struct regs *r);

/* IRQ.C */
void irq_handler(struct regs *r);
void irq_install();
void irq_remap(void);

/* TIMER.C */
extern void timer_install();
extern void timer_wait(int ticks);
extern void timer_handler(struct regs *r);
extern void timer_phase(int hz);

/* KB.C */
void keyboard_install();
void keyboard_handler(struct regs *r);

/* KHEAP.C */
extern u32int kmalloc_int(u32int sz, int align, u32int *phys);
extern u32int kmalloc_a(u32int sz);
extern u32int kmalloc_p(u32int sz, u32int *phys);
extern u32int kmalloc_ap(u32int sz, u32int *phys);
extern u32int kmalloc(u32int sz);

/* PAGE.C */

typedef struct page
{
    u32int present    : 1;   //0 Page present in memory
    u32int rw         : 1;   //1 Read-only if clear, readwrite if set
    u32int user       : 1;   //2 Supervisor level only if clear
    u32int intelRes1  : 2;   //3,4  
    u32int accessed   : 1;   // 5Has the page been accessed since last refresh?
    u32int intelRes2  : 3;   //6,7,8
    u32int sys      : 3;   //9,10,11 Has the page been written to since last refresh?
    //u32int unused     : 7;   // Amalgamation of unused and reserved bits
    u32int frame      : 20;  // Frame address (shifted right 12 bits)
} page_t;

typedef struct page_table
{
    page_t pages[1024];
} page_table_t;

typedef struct page_directory
{
    /**
       Array of pointers to pagetables.
    **/
    page_table_t *tables[1024];
    /**
       Array of pointers to the pagetables above, but gives their *physical*
       location, for loading into the CR3 register.
    **/
    u32int tablesPhysical[1024];

    /**
       The physical address of tablesPhysical. This comes into play
       when we get our kernel heap allocated and the directory
       may be in a different location in virtual memory.
    **/
    u32int physicalAddr;
} page_directory_t;

/**
   Sets up the environment, page directories etc and
   enables paging.
**/
extern void initialise_paging();

/**
   Causes the specified page directory to be loaded into the
   CR3 register.
**/
extern void switch_page_directory(page_directory_t *new);

/**
   Retrieves a pointer to the page required.
   If make == 1, if the page-table in which this page should
   reside isn't created, create it!
**/
extern page_t *get_page(u32int address, int make, page_directory_t *dir);

/**
   Handler for page faults.
**/
extern void page_fault(struct regs *r);

#endif
